import typing as T
import logging
from enum import Enum

FORMAT_STR = "%(asctime)s | %(process)d | %(name)s |  %(levelname)s: %(message)s"


class Color(str, Enum):
    BLACK = 0
    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    PURPLE = 5
    TEAL = 6
    GREY = 7


RESET_STR = "\x1b[0m"


def format_str_from_color(color: Color, bold: bool = False) -> str:
    bold_str = "20" if not bold else "1"
    color_str = f"\x1b[3{color.value};{bold_str}m"
    format_str = f"{color_str} {FORMAT_STR} {RESET_STR}"
    return format_str


COLORS = dict[int, Color]
FORMATS = dict[int, str]

DEFAULT_FORMATS: FORMATS = {
    logging.DEBUG: format_str_from_color(Color.TEAL),
    logging.INFO: format_str_from_color(Color.PURPLE),
    logging.WARNING: format_str_from_color(Color.YELLOW),
    logging.ERROR: format_str_from_color(Color.RED, bold=False),
    logging.CRITICAL: format_str_from_color(Color.RED, bold=True),
}


def formats_from_colors(color_map: COLORS) -> FORMATS:
    return {k: format_str_from_color(v) for k, v in color_map.items()}


def formats_from_color(color: Color) -> FORMATS:
    format_str = format_str_from_color(color)
    return {k: format_str for k in DEFAULT_FORMATS.keys()}


class CustomFormatter(logging.Formatter):
    def __init__(
        self,
        formats: FORMATS | None = None,
        no_colors: bool = False,
        *args: T.Any,
        **kwargs: T.Any,
    ):
        super().__init__(*args, **kwargs)
        self.formats = formats or DEFAULT_FORMATS
        self.no_colors = no_colors

    def format(self, record: logging.LogRecord) -> str:
        if self.no_colors:
            format_str = FORMAT_STR
        else:
            format_str = self.formats[record.levelno]
        formatter = logging.Formatter(format_str)
        return formatter.format(record)


def create_logger(
    name: str,
    level: T.Optional[T.Any] = logging.DEBUG,
    colors_map: T.Optional[COLORS] = None,
    no_colors: T.Optional[bool] = False,
    one_color: T.Optional[Color] = None,
) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level=level)
    handler = logging.StreamHandler()
    if one_color:
        formats = formats_from_color(one_color)
    elif colors_map:
        formats = formats_from_colors(colors_map)
    else:
        formats = None
    handler.setFormatter(CustomFormatter(formats=formats, no_colors=no_colors))
    logger.addHandler(handler)
    return logger


__all__ = ["FORMATS", "Color", "create_logger"]



import typing as T
import uuid
from dataclasses import dataclass, field
from enum import Enum
import graphql
from pydantic import BaseModel
from pydantic.fields import FieldInfo


class OperationType(str, Enum):
    query = "query"
    mutation = "mutation"


@dataclass
class Argument:
    name: str
    display_name: str
    value: T.Any | None


@dataclass(frozen=True)
class Node:
    id: uuid.UUID
    original_node: graphql.Node
    children: list[T.Union["FieldNode", "InlineFragmentNode"]] | None

    def __hash__(self):
        return hash(self.id)


@dataclass(frozen=True)
class FieldNode(Node):
    name: str
    alias: str | None
    display_name: str
    arguments: list[Argument]
    annotation: T.Any

    def __hash__(self):
        return hash(self.id)


@dataclass(frozen=True)
class FieldNodeField(FieldNode):
    field: FieldInfo

    def __hash__(self):
        return hash(self.id)


@dataclass(frozen=True)
class FieldNodeMethod(FieldNode):
    method: T.Callable

    def __hash__(self):
        return hash(self.id)


@dataclass(frozen=True)
class FieldNodeModel(FieldNode):
    models: list[T.Type[BaseModel]]

    def __hash__(self):
        return hash(self.id)


@dataclass(frozen=True)
class InlineFragmentNode(Node):
    type_condition: str
    annotation: T.Any

    def __hash__(self):
        return hash(self.id)


@dataclass(frozen=True)
class OperationNode(Node):
    name: str | None
    type: OperationType

    def __hash__(self):
        return hash(self.id)


__all__ = [
    "OperationType",
    "Argument",
    "Node",
    "FieldNode",
    "FieldNodeField",
    "FieldNodeModel",
    "FieldNodeMethod",
    "InlineFragmentNode",
    "OperationNode",
]



import random
import re
import string
import typing as T
from enum import Enum

import sqlparse
from pydantic import BaseModel, Field


class PostgresDriver(str, Enum):
    SQLALCHEMY = "SQLALCHEMY"
    PSYCOPG3 = "PSYCOPG3"
    ASYNCPG = "ASYNCPG"


ListT = T.TypeVar("ListT")


def random_string(length: int) -> str:
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


def chunk_list(lst: list[ListT], chunk_size: int) -> list[list[ListT]]:
    return [lst[i : i + chunk_size] for i in range(0, len(lst), chunk_size)]


def split_text_around_where(text: str) -> tuple[str, str | None]:
    # Regex pattern to split the string around 'where' (case-insensitive)
    pattern = r"(?i)\bwhere\b"

    # Splitting the string
    parts = re.split(pattern, text, maxsplit=1)

    # Stripping whitespace from both parts
    before_where = parts[0].strip()
    after_where = parts[1].strip() if len(parts) > 1 else None

    return before_where, after_where


class FilterConnector(str, Enum):
    AND = "AND"
    OR = "OR"


class QueryBuilderError(Exception):
    pass


class Cardinality(str, Enum):
    ONE = "ONE"
    MANY = "MANY"


class Selection(BaseModel):
    name: str

    @property
    def is_simple_column(self) -> bool:
        if not isinstance(self, SelectionField):
            return False
        return (
            self.name == self.path
            or self.path == f"$current.{self.name}"
            or self.path == f'"{self.path}"'
            or self.path == f'$current."{self.name}"'
        )


class SelectionField(Selection):
    path: str
    variables: dict[str, T.Any] | None = None


class SelectionSub(Selection):
    qb: "QueryBuilder"


class CTE(BaseModel):
    cte_str: str
    join_str: str
    is_top_level: bool = True


class QueryBuilder(BaseModel):
    table_name: str
    table_alias: str | None = None

    cardinality: Cardinality
    selections: list[SelectionField | SelectionSub] = Field(default_factory=list)
    variables: dict[str, T.Any] = Field(default_factory=dict)

    ctes: list[CTE] = Field(default_factory=list)

    from_: str | None = None
    where: str | None = None

    order_by: str | None = None
    offset: str | None = None
    limit: str | None = None

    full_query_str: str | None = None
    pattern_to_replace: str | None = None

    is_count: bool = False

    @staticmethod
    def build_child_var_name(
        child_name: str, var_name: str, variables_to_use: dict[str, T.Any]
    ) -> str:
        child_name = re.sub(r"[^a-zA-Z0-9]+", "_", child_name)
        count = 0
        while var_name in variables_to_use:
            count_str = "" if not count else f"_{count}"
            var_name = f"_{child_name}{count_str}_{var_name}"
        return var_name

    def add_variable(
        self, key: str, val: T.Any, replace: bool = False
    ) -> "QueryBuilder":
        if key in self.variables:
            if not replace and self.variables[key] != val:
                raise QueryBuilderError(
                    f"Key {key} already exists in variables so you cannot add it. "
                    f"If you'd like to replace it, pass replace."
                )
        self.variables[key] = val
        return self

    def add_variables(
        self, variables: dict[str, T.Any], replace: bool = False
    ) -> "QueryBuilder":
        """if there is an error, it does not save to the builder"""
        if not variables:
            return self
        if not replace:
            for k, v in variables.items():
                if k in self.variables and v != self.variables[k]:
                    raise QueryBuilderError(
                        f"Key '{k}' already exists in variables so you cannot add it. "
                        f"If you'd like to replace it, pass replace."
                    )
        self.variables.update(variables)
        return self

    def set_offset(self, offset: int | None, replace: bool = False) -> "QueryBuilder":
        if offset is None:
            return self
        if self.offset is not None:
            if not replace:
                raise QueryBuilderError(
                    "An offset already exists. If you would like to replace it, pass in replace."
                )
        self.offset = "OFFSET $offset"
        self.add_variable("offset", offset, replace=replace)
        return self

    def set_limit(self, limit: int | None, replace: bool = False) -> "QueryBuilder":
        if limit is None:
            return self
        if self.limit is not None:
            if not replace:
                raise QueryBuilderError(
                    "A limit already exists. If you would like to replace it, pass in replace."
                )
        self.limit = "LIMIT $limit"
        self.add_variable("limit", limit, replace=replace)
        return self

    def add_cte(
        self, cte_str: str, join_str: str, variables: dict[str, T.Any] | None = None
    ) -> "QueryBuilder":
        self.add_variables(variables)
        self.ctes.append(CTE(cte_str=cte_str, join_str=join_str))
        return self

    def set_from(
        self,
        from_: str,
        variables: dict[str, T.Any] | None = None,
        replace_from: bool = False,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        if self.from_ and not replace_from:
            raise QueryBuilderError("from_ already exists.")
        self.add_variables(variables=variables, replace=replace_variables)
        pre_where, post_where = split_text_around_where(from_)
        self.from_ = pre_where
        if post_where:
            self.and_where(post_where)
        return self

    def and_where(
        self,
        where: str,
        variables: dict[str, T.Any] | None = None,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        self.add_variables(variables=variables, replace=replace_variables)
        if not self.where:
            self.where = where
        else:
            self.where = f"{self.where} AND {where}"
        return self

    def set_where(
        self,
        where: str,
        variables: dict[str, T.Any] | None = None,
        replace_where: bool = False,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        if self.where and not replace_where:
            raise QueryBuilderError("Where by already exists.")
        self.add_variables(variables=variables, replace=replace_variables)
        self.where = where
        return self

    def set_order_by(
        self,
        order_by: str,
        variables: dict[str, T.Any] | None = None,
        replace_order_by: bool = False,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        if self.order_by and not replace_order_by:
            raise QueryBuilderError("Order by already exists.")
        self.add_variables(variables=variables, replace=replace_variables)
        self.order_by = order_by
        return self

    def set_full_query_str(
        self,
        full_query_str: str,
        replace: bool = False,
        variables: dict[str, T.Any] | None = None,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        if self.full_query_str and not replace:
            raise QueryBuilderError("full_query_str already exists.")
        self.add_variables(variables=variables, replace=replace_variables)
        pattern_to_replace = "".join(
            random.choice(string.ascii_letters + string.digits) for _ in range(10)
        )
        self.full_query_str = full_query_str.replace("$$", pattern_to_replace)
        self.pattern_to_replace = pattern_to_replace
        return self

    def build_subquery(
        self,
        name: str,
        qb: "QueryBuilder",
        path: tuple[str, ...],
        parent_table_alias: str,
        variables: dict,
        order_fields_alphabetically: bool,
    ) -> str:
        s, v = qb.build(
            parent_table_alias=parent_table_alias,
            path=path,
            order_fields_alphabetically=order_fields_alphabetically,
        )
        for var_name, var_val in v.items():
            if var_name in variables:
                if var_val is variables[var_name]:
                    continue
                # must change the name for the child
                new_var_name = self.build_child_var_name(
                    child_name=name,
                    var_name=var_name,
                    variables_to_use=variables,
                )
                variables[new_var_name] = var_val
                # now, must regex the str to find this and replace it
                regex = re.compile(rf"\${var_name}(?!\w)")
                s = regex.sub(f"${new_var_name}", s)
            else:
                variables[var_name] = var_val

        s = f"'{name}', ({s})"
        return s

    def build_fields_s(
        self,
        new_path: tuple[str, ...],
        table_alias: str,
        order_fields_alphabetically: bool,
    ) -> tuple[list[str], dict[str, T.Any]]:
        variables = self.variables.copy()
        subquery_strs: list[str] = [
            self.build_subquery(
                name=sel_sub.name,
                qb=sel_sub.qb,
                path=new_path,
                variables=variables,
                parent_table_alias=table_alias,
                order_fields_alphabetically=order_fields_alphabetically,
            )
            for sel_sub in self.selections
            if isinstance(sel_sub, SelectionSub)
        ]
        selection_strs: list[str] = [
            f"'{sel.name}', {sel.path}"
            for sel in self.selections
            if isinstance(sel, SelectionField)
        ]
        all_fields_strs = [*selection_strs, *subquery_strs]
        if order_fields_alphabetically:
            all_fields_strs.sort()
        if not all_fields_strs:
            raise QueryBuilderError(f"Query Builder {self=} has no fields.")

        return all_fields_strs, variables

    def build_filter_parts_s(self) -> str:
        filter_parts: list[str] = []
        if self.where:
            filter_parts.append(f"WHERE {self.where}")
        if self.order_by:
            filter_parts.append(f"ORDER BY {self.order_by}")
        if self.offset:
            filter_parts.append(self.offset)
        if self.limit:
            filter_parts.append(self.limit)
        filter_parts_s = "\n".join(filter_parts)
        return filter_parts_s

    @staticmethod
    def replace_current_and_parent(
        s: str, table_alias: str, parent_table_alias: str | None
    ) -> str:
        s = s.replace("$current", table_alias)
        if parent_table_alias:
            s = s.replace("$parent", parent_table_alias)
        return s

    def get_top_level_ctes(self) -> list[CTE]:
        ctes = [cte for cte in self.ctes if cte.is_top_level]
        for sel_sub in self.selections:
            if isinstance(sel_sub, SelectionSub):
                ctes.extend(sel_sub.qb.get_top_level_ctes())
        return ctes

    def build(
        self,
        parent_table_alias: str | None,
        path: tuple[str, ...] | None,
        order_fields_alphabetically: bool = True,
        is_count: bool | None = None,
        use_top_level_ctes: bool = True,
    ) -> tuple[str, dict[str, T.Any]]:
        is_count = is_count if is_count is not None else self.is_count
        if is_count:
            if self.limit is not None:
                raise QueryBuilderError("Cannot be is_count and have a limit.")
            if self.offset is not None:
                raise QueryBuilderError("Cannot be is_count and have an offset.")
        if path:
            new_path = (*path, self.table_name)
        else:
            new_path = (self.table_name,)
        if self.table_alias:
            table_alias = self.table_alias
        else:
            table_alias = "__".join(new_path).replace('"', "").replace(".", "__")
            if len(table_alias) > 55:
                table_alias = (
                    f"{table_alias[0:10]}{random_string(10)}{table_alias[-10:]}"
                )
        if not path:
            if table_alias.lower() == self.table_name.lower().replace('"', ""):
                table_alias = f"_{table_alias}"
        strs_list, variables = self.build_fields_s(
            new_path=new_path,
            table_alias=table_alias,
            order_fields_alphabetically=order_fields_alphabetically,
        )
        filter_parts_s = self.build_filter_parts_s()
        # now do from_
        if not self.from_:
            self.from_ = "*FROM*"
        self.from_ = self.from_.replace(
            "*FROM*", f"FROM {self.table_name} {table_alias}"
        )
        if not self.from_ or not self.from_.lower().startswith("from "):
            self.from_ = f"FROM {self.table_name} {table_alias} {self.from_}"

        # if this is the top level, use top level CTES. Otherwise, ignore them
        if use_top_level_ctes:
            ctes_to_use = self.get_top_level_ctes()
        else:
            ctes_to_use = [cte for cte in self.ctes if not cte.is_top_level]

        cte_str = ",\n".join([cte.cte_str for cte in ctes_to_use])
        cte_join_str = "\n".join([cte.join_str for cte in ctes_to_use])

        # now build the json objects by chunking the keys + vals at 50
        chunks = chunk_list(lst=strs_list, chunk_size=50)
        if len(chunks) == 1:
            json_obj_str = f'json_build_object({", ".join(chunks[0])})'
        else:
            json_obj_strs: list[str] = []
            for chunk in chunks:
                json_obj_strs.append(f'jsonb_build_object({", ".join(chunk)})')
            json_obj_str = " || ".join(json_obj_strs)
        if is_count:
            json_obj_str = f"COUNT({json_obj_str})"
        s = f"""
{cte_str}
SELECT {json_obj_str} AS {table_alias}_json
{self.from_}
{cte_join_str}
{filter_parts_s}
""".strip()
        if self.cardinality == Cardinality.MANY:
            if is_count is not True:
                s = f"""
    SELECT COALESCE(json_agg({table_alias}_json), '[]'::json) AS {table_alias}_json_agg
    FROM (
        {s}
    ) as {table_alias}_json_sub
                """.strip()
        if self.full_query_str:
            s = self.full_query_str.replace(self.pattern_to_replace, s)
        # now replace the values
        s = self.replace_current_and_parent(
            s=s, table_alias=table_alias, parent_table_alias=parent_table_alias
        )
        return s, variables

    def build_root(
        self,
        format_sql: bool = False,
        order_fields_alphabetically: bool = False,
        parent_table_alias: str = None,
        path: tuple[str, ...] = None,
        driver: PostgresDriver = PostgresDriver.SQLALCHEMY,
        is_count: bool | None = None,
    ) -> tuple[str, dict[T.Any]]:
        rr = self.build(
            order_fields_alphabetically=order_fields_alphabetically,
            parent_table_alias=parent_table_alias,
            path=path,
            is_count=is_count,
        )
        s, v = rr
        if v:
            if driver == PostgresDriver.SQLALCHEMY:
                s, variables = self.prepare_query_sqlalchemy(sql=s, params=v)
            elif driver == PostgresDriver.PSYCOPG3:
                s, variables = self.prepare_query_psycopg(sql=s, params=v)
            elif driver == PostgresDriver.ASYNCPG:
                s, variables = self.prepare_query_asyncpg(sql=s, params=v)
            else:
                raise QueryBuilderError(f"Unknown driver: {driver=}.")
        else:
            variables = {}
        if format_sql:
            s = sqlparse.format(s, reindent=True, keyword_case="upper")
        return s, variables

    @staticmethod
    def prepare_query_asyncpg(
        sql: str, params: dict[str, T.Any]
    ) -> tuple[str, list[T.Any]]:
        """
        Generated by GPT4
        Converts a SQL string with named parameters (e.g., $variable) to a format
        compatible with asyncpg (using $1, $2, etc.), and returns the new SQL string
        and the list of values in the correct order.

        :param sql: Original SQL string with named parameters
        :param params: Dictionary of parameters
        :return: Tuple of (new_sql_string, list_of_values)
        """

        # Extract the named parameters from the SQL string
        named_params = re.findall(r"\$(\w+)", sql)
        # Ensure that each parameter is unique
        unique_params = list(dict.fromkeys(named_params))

        # Replace named parameters with positional parameters ($1, $2, etc.)
        for i, param in enumerate(unique_params, start=1):
            sql = sql.replace(f"${param}", f"${i}")

        # Create the list of values in the order they appear in the query
        values = [params[param] for param in unique_params]

        return sql, values

    @staticmethod
    def prepare_query_psycopg(
        sql: str, params: dict[str, T.Any]
    ) -> tuple[str, dict[str, T.Any]]:
        """
        Converts a SQL string with named parameters (e.g., $variable) to a format
        compatible with psycopg (using %(variable)s), and returns the new SQL string
        and the dictionary of parameters.

        :param sql: Original SQL string with named parameters
        :param params: Dictionary of parameters
        :return: Tuple of (new_sql_string, dict_of_parameters)
        """

        # Extract the named parameters from the SQL string
        named_params = re.findall(r"[\$:]([\w]+)", sql)

        # Replace named parameters with %(param)s
        for param in named_params:
            sql = sql.replace(f"${param}", f"%({param})s")
            # TODO be careful with this lower one
            sql = sql.replace(f":{param}", f"%({param})s")
        return sql, {**params}

    @staticmethod
    def prepare_query_sqlalchemy(
        sql: str, params: dict[str, T.Any]
    ) -> tuple[str, dict[str, T.Any]]:
        """
        Converts a SQL string with named parameters (e.g., $variable) to a format
        compatible with sqlalchemy (using :variable), and returns the new SQL string
        and the dictionary of parameters.

        :param sql: Original SQL string with named parameters
        :param params: Dictionary of parameters
        :return: Tuple of (new_sql_string, dict_of_parameters)
        """

        # Extract the named parameters from the SQL string
        named_params = re.findall(r"\$(\w+)", sql)

        # Replace named parameters with %(param)s
        for param in named_params:
            sql = sql.replace(f"${param}", f":{param}")

        # Create the dictionary of parameters to be used in the query
        return sql, {**params}

    def existing_sel(self, name: str) -> SelectionField | SelectionSub | None:
        for sel in self.selections:
            if name == sel.name:
                return sel
        return None

    def sel(
        self, name: str, path: str = None, variables: dict[str, T.Any] | None = None
    ) -> "QueryBuilder":
        if not path:
            path = f'$current."{name}"'
        self.add_variables(variables)
        self.selections.append(
            SelectionField(name=name, path=path, variables=variables)
        )
        return self

    def sel_sub(self, name: str, qb: "QueryBuilder") -> "QueryBuilder":
        self.selections.append(SelectionSub(name=name, qb=qb))
        return self

    def add_sel(self, sel: SelectionField | SelectionSub) -> "QueryBuilder":
        if isinstance(sel, SelectionField):
            return self.sel(name=sel.name, path=sel.path, variables=sel.variables)
        elif isinstance(sel, SelectionSub):
            return self.sel_sub(name=sel.name, qb=sel.qb)
        else:
            raise QueryBuilderError(
                f"Can only add selections for SelectionField or SelectionSub, not {type(sel)}"
            )



import typing as T
import uuid
import graphql
from graphql.type.definition import GraphQLNullableType
from .models import (
    FieldNode,
    FieldNodeModel,
    FieldNodeMethod,
    FieldNodeField,
    OperationNode,
    OperationType,
    InlineFragmentNode,
    Argument,
)


class Translator:
    def __init__(
        self,
        document: graphql.DocumentNode,
        schema: graphql.GraphQLSchema,
        display_to_python_map: dict[str, str],
    ):
        self.document = document
        self.schema = schema
        self.display_to_python_map = display_to_python_map
        self.query_definitions: list[graphql.OperationDefinitionNode] = []
        self.mutation_definitions: list[graphql.OperationDefinitionNode] = []
        self.subscription_definitions: list[graphql.OperationDefinitionNode] = []
        self.fragment_definitions: dict[str, graphql.FragmentDefinitionNode] = {}

    def parse_val(self, val: graphql.ValueNode) -> T.Any:
        if isinstance(val, graphql.VariableNode):
            return val
        if isinstance(val, graphql.IntValueNode):
            return int(val.value)
        if isinstance(val, graphql.FloatValueNode):
            return float(val.value)
        if isinstance(val, graphql.StringValueNode):
            return val.value
        if isinstance(val, graphql.BooleanValueNode):
            return bool(val.value)
        if isinstance(val, graphql.ObjectValueNode):
            return {
                field.name.value: self.parse_val(field.value) for field in val.fields
            }
        if isinstance(val, graphql.ListValueNode):
            return [self.parse_val(v) for v in val.values]
        if isinstance(val, graphql.NullValueNode):
            return None
        # TODO for enums, lists... might need to use pydantic validate call here...anoying w return types and all
        return val.value

    @staticmethod
    def combine_sels(
        a: graphql.InlineFragmentNode | graphql.FieldNode,
        b: graphql.InlineFragmentNode | graphql.FieldNode,
    ) -> graphql.InlineFragmentNode | graphql.FieldNode:
        new_node: graphql.InlineFragmentNode | graphql.FieldNode = a.__copy__()
        if isinstance(a, graphql.FieldNode):
            new_node.arguments = (*a.arguments, *b.arguments)
        if not new_node.selection_set:
            new_node.selection_set = graphql.SelectionNode()
        new_node.selection_set.selections = (
            *(a.selection_set.selections if a.selection_set else []),
            *(b.selection_set.selections if b.selection_set else []),
        )
        return new_node

    def children_from_node(
        self,
        gql_field: graphql.GraphQLField | graphql.GraphQLObjectType,
        node: graphql.FieldNode | graphql.InlineFragmentNode,
        path_to_children: tuple[str, ...],
    ) -> list[FieldNode | InlineFragmentNode] | None:
        if node.selection_set is None:
            return None
        children: list[FieldNode | InlineFragmentNode] = []
        root_field = self.get_root_type(type_=gql_field)
        selection_q = list({*node.selection_set.selections})

        # first, flatten and combine them
        has_seen: dict[str, graphql.InlineFragmentNode | graphql.FieldNode] = {}
        while len(selection_q) > 0:
            sel = selection_q.pop(0)
            if isinstance(sel, graphql.FragmentSpreadNode):
                frag = self.fragment_definitions[sel.name.value]
                selection_q.extend(frag.selection_set.selections)
                continue
            if isinstance(sel, graphql.FieldNode):
                key = sel.alias.value if sel.alias else sel.name.value
            elif isinstance(sel, graphql.InlineFragmentNode):
                key = sel.type_condition.name
            else:
                raise Exception(f"Invalid sel: {sel=}")
            if existing := has_seen.get(key):
                sel = self.combine_sels(existing, sel)
            has_seen[key] = sel

        for sel in has_seen.values():
            if isinstance(sel, graphql.InlineFragmentNode):
                gql_field = self.get_root_type(
                    root_field, type_condition=sel.type_condition.name.value
                )
            else:
                sel_name = sel.name.value
                if sel_name == "__typename":
                    gql_field = None
                else:
                    gql_field = root_field.fields[sel_name]
            child_s = self.from_node(
                gql_field=gql_field, node=sel, path=path_to_children
            )
            children.extend(child_s if isinstance(child_s, list) else [child_s])
        return children

    # @cache #TODO cache but these models are not hashable...
    def get_root_type(
        self,
        type_: T.Union[
            GraphQLNullableType,
            graphql.GraphQLNonNull,
            graphql.GraphQLField,
            graphql.GraphQLUnionType,
        ],
        type_condition: str = None,
    ) -> graphql.GraphQLObjectType | tuple[graphql.GraphQLUnionType] | None:
        if hasattr(type_, "of_type"):
            return self.get_root_type(type_.of_type, type_condition=type_condition)
        if hasattr(type_, "type"):
            return self.get_root_type(type_.type, type_condition=type_condition)
        if type_condition:
            if hasattr(type_, "types"):
                types = [self.get_root_type(t) for t in type_.types]
                for t in types:
                    if t.name == type_condition:
                        return t
                raise Exception("Type condition was not found.")

        return type_

    def from_node(
        self,
        gql_field: graphql.GraphQLField | graphql.GraphQLObjectType | None,
        node: graphql.SelectionNode,
        path: tuple[str, ...],
    ) -> FieldNode | InlineFragmentNode | list[FieldNode | InlineFragmentNode]:
        """if this is a fragment, return many nodes"""
        if isinstance(node, graphql.FragmentSpreadNode):
            frag = self.fragment_definitions[node.name.value]
            return [
                self.from_node(
                    gql_field=gql_field,
                    node=sel,
                    path=path,
                )
                for sel in frag.selection_set.selections
            ]

        if not gql_field:
            annotation = None
        elif isinstance(gql_field, graphql.GraphQLObjectType):
            annotation = gql_field._pydantic_model
        else:
            annotation = gql_field.type._anno

        if isinstance(node, graphql.InlineFragmentNode):
            type_condition = node.type_condition.name.value
            return InlineFragmentNode(
                id=uuid.uuid4(),
                children=self.children_from_node(
                    gql_field=gql_field,
                    node=node,
                    path_to_children=(*path, type_condition),
                ),
                original_node=node,
                type_condition=type_condition,
                annotation=annotation,
            )
        elif isinstance(node, graphql.FieldNode):
            # build args
            arguments: list[Argument] = []
            for argument in node.arguments:
                arguments.append(
                    Argument(
                        display_name=argument.name.value,
                        name=self.display_to_python_map[argument.name.value],
                        value=self.parse_val(argument.value),
                    )
                )
            alias = node.alias.value if node.alias else None
            try:
                name = self.display_to_python_map[node.name.value]
            except KeyError as e:
                if node.name.value == "__typename":
                    name = node.name.value
                else:
                    raise e

            children = self.children_from_node(
                gql_field=gql_field,
                node=node,
                path_to_children=(*path, node.name.value),
            )
            display_name = node.name.value
            # gql_field_type = self.get_root_type(gql_field)
            if not gql_field:
                # this is __typename
                return FieldNode(
                    id=uuid.uuid4(),
                    original_node=node,
                    children=children,
                    name=name,
                    alias=alias,
                    display_name=display_name,
                    arguments=arguments,
                    annotation=annotation,
                )
            gql_field_type = gql_field.type
            if hasattr(gql_field_type, "_field_info"):
                return FieldNodeField(
                    id=uuid.uuid4(),
                    original_node=node,
                    children=children,
                    name=name,
                    alias=alias,
                    display_name=display_name,
                    arguments=arguments,
                    annotation=annotation,
                    field=gql_field_type._field_info,
                )
            elif hasattr(gql_field_type, "_method"):
                return FieldNodeMethod(
                    id=uuid.uuid4(),
                    original_node=node,
                    children=children,
                    name=name,
                    alias=alias,
                    display_name=display_name,
                    arguments=arguments,
                    annotation=annotation,
                    method=gql_field_type._method,
                )
            else:
                root_type = self.get_root_type(gql_field)
                if isinstance(root_type, graphql.GraphQLUnionType):
                    models = [t._pydantic_model for t in root_type.types]
                elif hasattr(root_type, "_pydantic_model"):
                    models = [root_type._pydantic_model]
                else:
                    # hacky fix because the enum wasn't being registered
                    if isinstance(root_type, graphql.GraphQLEnumType):
                        gql_field_temp = getattr(gql_field_type, "of_type")
                        while not hasattr(gql_field_temp, "_field_info"):
                            # TODO not sure if this is okay
                            if not hasattr(gql_field_temp, "of_type"):
                                return FieldNodeField(
                                    id=uuid.uuid4(),
                                    original_node=node,
                                    children=children,
                                    name=name,
                                    alias=alias,
                                    display_name=display_name,
                                    arguments=arguments,
                                    annotation=annotation,
                                    field=None,
                                )
                            gql_field_temp = getattr(gql_field_temp, "of_type")
                        return FieldNodeField(
                            id=uuid.uuid4(),
                            original_node=node,
                            children=children,
                            name=name,
                            alias=alias,
                            display_name=display_name,
                            arguments=arguments,
                            annotation=annotation,
                            field=gql_field_temp._field_info,
                        )
                # otherwise, this will be an object type
                return FieldNodeModel(
                    id=uuid.uuid4(),
                    original_node=node,
                    children=children,
                    name=name,
                    alias=alias,
                    display_name=display_name,
                    arguments=arguments,
                    models=models,
                    annotation=annotation,
                )
        else:
            raise Exception(f"Invalid node type: {type(node)=}, {node=}")

    def from_operation_node(
        self, node: graphql.OperationDefinitionNode
    ) -> OperationNode:
        # first, build children, and then their children
        children: list[FieldNode | InlineFragmentNode] = []
        # TODO possible inline frags?
        for sel in node.selection_set.selections:
            if node.operation == graphql.OperationType.MUTATION:
                gql_field = self.schema.mutation_type
            elif node.operation == graphql.OperationType.QUERY:
                gql_field = self.schema.query_type
            else:
                raise Exception(f"Unimplemented operation type: {node.operation=}")
            child_s = self.from_node(
                gql_field=gql_field.fields[sel.name.value],
                node=sel,
                path=(node.operation.value,),
            )
            if type(child_s) is list:
                children.extend(child_s)
            else:
                children.append(child_s)

        op = OperationNode(
            id=uuid.uuid4(),
            name=node.name.value if node.name else None,
            type=OperationType(node.operation.value),
            children=children,
            original_node=node,
        )

        return op

    def translate(self) -> list[OperationNode]:
        for definition in self.document.definitions:
            if isinstance(definition, graphql.FragmentDefinitionNode):
                self.fragment_definitions[definition.name.value] = definition
            elif isinstance(definition, graphql.OperationDefinitionNode):
                if definition.operation == graphql.OperationType.QUERY:
                    self.query_definitions.append(definition)
                elif definition.operation == graphql.OperationType.MUTATION:
                    self.mutation_definitions.append(definition)
                else:
                    self.subscription_definitions.append(definition)
            else:
                raise Exception(
                    f"Unknown type of definition: {definition=}, {type(definition)=}"
                )

        # start with the operations
        op_nodes: list[OperationNode] = []
        for operation in [*self.query_definitions, *self.mutation_definitions]:
            op_nodes.append(self.from_operation_node(operation))
        return op_nodes



import typing as T
import traceback
import asyncio
import inspect

from fastapi import Request, Response, BackgroundTasks
from pydantic import TypeAdapter, ValidationError

from fastgql.gql_ast import models as M
from fastgql.gql_models import GQL, GQLError
from fastgql.depends import Depends
from fastgql.execute.utils import InfoType, parse_value, Info, ContextType
from fastgql.utils import node_from_path
import pydantic_core


class Resolver:
    def __init__(
        self,
        *,
        operation_name: str | None,
        display_to_python_map: dict[str, str],
        info_cls: T.Type[InfoType],
        context_cls: T.Type[ContextType],
        is_not_nullable_map: dict[str, dict[str, bool]],
        variables: dict[str, T.Any] | None,
        request: Request,
        response: Response,
        bt: BackgroundTasks,
    ):
        self.operation_name = operation_name
        self.display_to_python_map = display_to_python_map
        self.info_cls = info_cls
        self.is_not_nullable_map = is_not_nullable_map
        self.variables = variables

        self.request = request
        self.response = response
        self.bt = bt

        self.errors: list[GQLError] = []

        self.context: ContextType = context_cls(
            request=self.request,
            response=self.response,
            background_tasks=self.bt,
            errors=self.errors,
            variables=self.variables,
        )

    async def inject_dependencies(
        self, func: T.Callable[..., T.Any], kwargs: dict[str, T.Any], info: InfoType
    ) -> T.Any:
        # if you need to build kwargs, do it. Then execute
        new_kwargs = await self.build_kwargs(func=func, kwargs=kwargs, info=info)
        res = func(**new_kwargs)
        if inspect.isawaitable(res):
            res = await res
        return res

    async def build_kwargs(
        self, func: T.Callable[..., T.Any], kwargs: dict[str, T.Any], info: InfoType
    ) -> dict[str, T.Any]:
        new_kwargs: dict[str, T.Any] = {}
        sig = inspect.signature(func)
        for name, param in sig.parameters.items():
            if name in kwargs:
                val = kwargs[name]
                if val is not None:
                    try:
                        val = TypeAdapter(param.annotation).validate_python(
                            val,
                            context={
                                "_display_to_python_map": self.display_to_python_map
                            },
                        )
                    except ValidationError as e:
                        if e.errors()[0]["type"] == "list_type" and not isinstance(
                            val, list
                        ):
                            val = TypeAdapter(param.annotation).validate_python(
                                [val],
                                context={
                                    "_display_to_python_map": self.display_to_python_map
                                },
                            )
                        else:
                            raise e
                new_kwargs[name] = val
            else:
                if isinstance(param.default, Depends):
                    new_kwargs[name] = await self.inject_dependencies(
                        func=param.default.dependency, kwargs=kwargs, info=info
                    )
                elif inspect.isclass(param.annotation):
                    if issubclass(param.annotation, Info):
                        new_kwargs[name] = info
                    elif issubclass(param.annotation, Request):
                        new_kwargs[name] = self.request
                    elif issubclass(param.annotation, Response):
                        new_kwargs[name] = self.response
                    elif issubclass(param.annotation, BackgroundTasks):
                        new_kwargs[name] = self.bt
                # just continue, the value is not given and that is okay
        return new_kwargs

    def get_info_key(self, func: T.Callable[..., T.Any]) -> str | None:
        sig = inspect.signature(func)
        for name, param in sig.parameters.items():
            if param.annotation == self.info_cls:
                return name

    async def inject_dependencies_and_execute(
        self,
        node: M.FieldNode,
        parent: M.Node,
        method: T.Callable | T.Awaitable,
        kwargs: dict[str, T.Any],
        new_path: tuple[str, ...],
    ) -> dict[str, T.Any] | list[dict[str, T.Any]] | T.Any | None:
        # TODO if inefficient, lazily create info! Or maybe even cache it... or come up with a better way
        info = self.info_cls(
            node=node,
            parent_node=parent,
            path=new_path,
            context=self.context,
        )
        new_kwargs = await self.build_kwargs(func=method, kwargs=kwargs, info=info)
        child_model_s = method(**new_kwargs)
        if inspect.isawaitable(child_model_s):
            child_model_s = await child_model_s
        if child_model_s:
            if node.children:
                return await self.resolve_node_s(
                    node=node, model_s=child_model_s, path=new_path
                )
            else:
                return child_model_s
        return child_model_s

    async def resolve_node_s(
        self,
        node: M.FieldNode | M.OperationNode,
        model_s: list[GQL] | GQL,
        path: tuple[str, ...],
    ) -> dict[str, T.Any] | None | list[dict[str, T.Any] | None]:
        if isinstance(model_s, list):
            return list(
                await asyncio.gather(
                    *[
                        self.resolve_node(node=node, model=model, path=path)
                        for model in model_s
                    ]
                )
            )
        return await self.resolve_node(node=node, model=model_s, path=path)

    async def resolve_node(
        self,
        node: M.FieldNode | M.OperationNode,
        model: GQL,
        path: tuple[str, ...],
    ) -> dict[str, T.Any] | None:
        if model is None:
            return None
        if not isinstance(model, GQL):
            raise Exception(
                f"Model {model.__class__.__name__} must be an instance of GQL."
            )
        # return final dict, or array of dicts, for the node
        final_d: dict[str, T.Any] = {}
        fields_to_include: dict[str, str] = {}
        name_to_return_to_display_name: dict[str, str] = {}
        children_q = [*node.children]
        proms_map: dict[str, T.Awaitable] = {}
        while len(children_q) > 0:
            child = children_q.pop(0)
            if isinstance(child, M.InlineFragmentNode):
                if child.type_condition == model.gql_type_name():
                    children_q[:0] = child.children
                continue
            name_to_return = child.alias or child.display_name
            name_to_return_to_display_name[name_to_return] = child.display_name
            if child in self.context.overwrite_return_value_map:
                final_d[name_to_return] = self.context.overwrite_return_value_map[child]
                continue
            new_path = (*path, name_to_return)
            if child.name == "__typename":
                final_d[name_to_return] = model.gql_type_name()
            else:
                if child.name not in model.model_fields:
                    # this must be a function

                    kwargs = {}
                    for arg in child.arguments:
                        _v = parse_value(variables=self.variables, v=arg.value)
                        if _v != pydantic_core.PydanticUndefined:
                            kwargs[arg.name] = _v

                    proms_map[name_to_return] = self.inject_dependencies_and_execute(
                        method=getattr(model, child.name),
                        node=child,
                        parent=node,
                        kwargs=kwargs,
                        new_path=new_path,
                    )
                else:
                    if not child.children:
                        # this is a property
                        fields_to_include[child.name] = name_to_return
                    else:
                        # this is either a BaseModel or a list of BaseModel
                        proms_map[name_to_return] = self.resolve_node_s(
                            node=child,
                            path=new_path,
                            model_s=getattr(model, child.name),
                        )

        # now gather the await-ables
        if proms_map:
            vals = await asyncio.gather(*proms_map.values(), return_exceptions=True)
            for name, val in zip(proms_map.keys(), vals):
                if isinstance(val, Exception):
                    if isinstance(val, GQLError):
                        self.errors.append(val)
                    else:
                        # do not expose this error to client
                        # include stack trace and send to sentry
                        node_that_errored = node_from_path(
                            node=node,
                            path=[name],
                            use_field_to_use=True,
                        )
                        self.errors.append(
                            GQLError(
                                message="Internal Server Error",
                                node=node_that_errored,
                                original_error=val,
                                path=(*path, name),
                            )
                        )
                        traceback.print_exception(val)
                    val = None
                final_d[name] = val
        model_d = model.model_dump(mode="json", include=set(fields_to_include))
        for name, name_to_use in fields_to_include.items():
            final_d[name_to_use] = model_d[name]
            if name_to_use != name:
                final_d[name] = model_d[name]

        # now null check and order property
        sorted_d = {}
        for name_to_return, name in name_to_return_to_display_name.items():
            val = final_d[name_to_return]
            if (
                val is None or isinstance(val, list) and None in val
            ):  # TODO speed test, must go thru every list?
                if self.is_not_nullable_map[model.gql_type_name()][name]:
                    # get the actual node from the path
                    null_node = node_from_path(
                        node=node, path=[name_to_return], use_field_to_use=True
                    )
                    if (
                        null_node not in self.context.overwrite_return_value_map
                        and path
                    ):
                        full_path_to_error = (*path, name_to_return)
                        self.errors.append(
                            GQLError(
                                message=f"Cannot return null for non-nullable field {'.'.join(full_path_to_error)}",
                                path=full_path_to_error,
                                node=null_node,
                            )
                        )
                    if node not in self.context.overwrite_return_value_map:
                        self.context.overwrite_return_value_map[node] = None
                    return None
            sorted_d[name_to_return] = val
        del final_d
        return sorted_d

    async def resolve_root_nodes(
        self,
        root_nodes: list[M.OperationNode],
        operation_type_to_model: dict[M.OperationType, GQL | None],
    ) -> dict[str, T.Any] | None:
        proms = []
        for root_node in root_nodes:
            model = operation_type_to_model[root_node.type]
            if not model:
                raise Exception(f"{root_node.type} type required.")
            proms.append(
                self.resolve_node(
                    node=root_node,
                    model=operation_type_to_model[root_node.type],
                    path=(),
                )
            )
        d_list = await asyncio.gather(*proms)
        final_d = {}
        for d in d_list:
            if d is None:
                return None
            final_d.update(d)
        return final_d



import typing as T
import dataclasses
from fastgql.gql_ast import models as M
from fastgql.context import BaseContext

ContextType = T.TypeVar("ContextType", bound=BaseContext)


@dataclasses.dataclass
class Info(T.Generic[ContextType]):
    """needed to make this a raw dataclass because context needs to be kept as a reference... pydantic copies dicts"""

    node: M.FieldNode | M.InlineFragmentNode
    parent_node: M.FieldNode | M.InlineFragmentNode | M.OperationNode
    path: tuple[str, ...]

    context: ContextType


__all__ = ["Info", "ContextType"]
